"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.listenerChain = exports.delay = exports.createEmitter = exports.createEffect = exports.configure = exports.getKey = exports.makeKey = exports.createFlow = exports.createController = exports.NO_KEY = exports.GLOBAL_STATUS_EVENT = exports.FLOW_UPDATE_EVENT = void 0;
exports.FLOW_UPDATE_EVENT = "#flow";
exports.GLOBAL_STATUS_EVENT = "#status";
exports.NO_KEY = {};
/**
 * create flow controller
 */
function createController(_a) {
    var _b = _a === void 0 ? {} : _a, context = _b.context, _c = _b.initData, initData = _c === void 0 ? {} : _c, initFlow = _b.initFlow, onData = _b.onData, onCompleted = _b.onCompleted, onFaulted = _b.onFaulted, _d = _b.gcInterval, gcInterval = _d === void 0 ? 60000 : _d;
    var data = __assign({}, initData);
    var promise = Promise.resolve();
    var ready = true;
    var error;
    var gcTimeout;
    var flows = new Map();
    var emitter = createEmitter({ wildcard: "*", privateEventPrefix: "#" });
    var controller = {
        context: context,
        get ready() {
            return ready;
        },
        get error() {
            return error;
        },
        get promise() {
            return promise;
        },
        emit: emitter.emit,
        on: emitter.on,
        start: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("start", args);
        },
        restart: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("restart", args);
        },
        cancelFlow: function (key) {
            var _a;
            (_a = flows.get(getKey(key))) === null || _a === void 0 ? void 0 : _a.cancel();
        },
        replaceFlow: function (currentFlow, withFlow) {
            if (flows.get(currentFlow.key) !== currentFlow) {
                return;
            }
            flows.set(currentFlow.key, withFlow);
        },
        flow: function (key, fn) {
            var keyType = typeof key;
            if (keyType === "function") {
                fn = key;
            }
            key = getKey(key);
            // flow(key)
            if (keyType !== "function" && arguments.length === 1) {
                return flows.get(key);
            }
            var flow = flows.get(key);
            if (!flow) {
                var data_1 = undefined;
                var status_1;
                var hasData = false;
                if (typeof key === "string" && key in initData) {
                    data_1 = initData[key];
                    hasData = true;
                    status_1 = "completed";
                }
                flow = createFlow({
                    controller: controller,
                    key: key,
                    fn: fn,
                    initData: data_1,
                    initStatus: status_1,
                    hasData: hasData,
                });
                flows.set(key, flow);
                if (hasData) {
                    emitter.emit(exports.FLOW_UPDATE_EVENT, flow);
                }
            }
            return flow;
        },
        remove: function (key) {
            var _a;
            key = getKey(key);
            // contains wildcard
            if (typeof key === "string" && key.indexOf("*") !== -1) {
                // ends with
                if (key[0] === "*") {
                    key = key.substr(1);
                    flows.forEach(function (flow, k) {
                        if (typeof k === "string" && k.endsWith(key)) {
                            flow.dispose();
                            flows.delete(k);
                        }
                    });
                }
                else {
                    // starts with
                    key = key.substr(0, key.length - 1);
                    flows.forEach(function (flow, k) {
                        if (typeof k === "string" && k.startsWith(key)) {
                            flow.dispose();
                            flows.delete(k);
                        }
                    });
                }
            }
            else {
                (_a = flows.get(key)) === null || _a === void 0 ? void 0 : _a.dispose();
                flows.delete(key);
            }
        },
        flowUpdated: function (flow) {
            var _a;
            if (flows.get(flow.key) !== flow)
                return;
            if (flow.status === "completed") {
                if (onData && typeof flow.key === "string") {
                    if (data[flow.key] !== flow.data) {
                        data = __assign(__assign({}, data), (_a = {}, _a[flow.key] = flow.data, _a));
                        onData(data);
                    }
                }
                onCompleted === null || onCompleted === void 0 ? void 0 : onCompleted(flow);
            }
            else if (flow.status === "faulted") {
                onFaulted === null || onFaulted === void 0 ? void 0 : onFaulted(flow);
            }
            emitter.emit(exports.FLOW_UPDATE_EVENT, flow);
        },
        run: function (fn) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            var cancel;
            return Object.assign(new Promise(function (resolve, reject) {
                var flow = createFlow({
                    controller: controller,
                    fn: fn,
                    key: exports.NO_KEY,
                    onEnd: function () {
                        if (flow.faulted) {
                            return reject(flow.error);
                        }
                        return resolve(flow.data);
                    },
                });
                cancel = flow.cancel;
                flow.start.apply(flow, args);
            }), {
                cancel: function () {
                    cancel === null || cancel === void 0 ? void 0 : cancel();
                },
            });
        },
        gc: function () {
            clearTimeout(gcTimeout);
            gc();
        },
    };
    function gc() {
        flows.forEach(function (flow, key) {
            if (flow.expiry && flow.expiry <= Date.now()) {
                flows.delete(key);
            }
        });
        if (gcInterval) {
            gcTimeout = setTimeout(gc, gcInterval);
        }
    }
    function run(type, inputs) {
        var _a;
        var args = [];
        var key = getKey(inputs[0]);
        var fn;
        // preload(flow, ...args)
        if (typeof inputs[0] === "function") {
            fn = inputs[0];
            args.push.apply(args, inputs.slice(1));
        }
        else {
            // preload(key, flow, ...args)
            fn = inputs[1];
            if (Array.isArray(inputs[0])) {
                args.push.apply(args, inputs[0].slice(1));
            }
            else {
                args.push.apply(args, inputs.slice(2));
            }
        }
        return (_a = controller.flow(key, fn))[type].apply(_a, args);
    }
    if (initFlow) {
        ready = false;
        promise = new Promise(function (resolve, reject) {
            createFlow({
                controller: controller,
                key: {},
                fn: initFlow,
                onSuccess: function (data) {
                    if (isPlainObject(data)) {
                        Object.assign(initData, data);
                    }
                    ready = true;
                    resolve();
                },
                onError: function (e) {
                    ready = true;
                    error = e;
                    reject(error);
                },
            }).start();
        });
    }
    gc();
    return controller;
}
exports.createController = createController;
function createFlow(_a) {
    var controller = _a.controller, parent = _a.parent, previous = _a.previous, key = _a.key, fn = _a.fn, onUpdate = _a.onUpdate, onSuccess = _a.onSuccess, onError = _a.onError, onEnd = _a.onEnd, initData = _a.initData, _b = _a.initStatus, initStatus = _b === void 0 ? "idle" : _b, hasData = _a.hasData, _c = _a.extra, extra = _c === void 0 ? {} : _c;
    var stale = 0;
    var status = initStatus;
    var error;
    var childError;
    var data = initData;
    var cancelled = false;
    var disposed = false;
    var called = 0;
    var currentNext;
    var cancelFn;
    var expiry = 0;
    var blocked = 0;
    var mergeFn;
    var iteratorStack = [];
    var emitter = createEmitter();
    function cleanup() {
        previous = undefined;
        onEnd === null || onEnd === void 0 ? void 0 : onEnd(data, error);
        emitter.emit("end");
        emitter.dispose();
    }
    function isExpired() {
        return expiry && expiry <= Date.now();
    }
    function isRunning() {
        return (!disposed &&
            !isStale(true) &&
            !isExpired() &&
            !flow.cancelled &&
            status === "running");
    }
    function notifyUpdate() {
        if (key === exports.NO_KEY) {
            return;
        }
        controller.flowUpdated(flow);
    }
    function statusChanged(newStatus, value, forceUpdate) {
        if (!forceUpdate && !isRunning())
            return;
        stale = 0;
        status = newStatus;
        cancelFn = undefined;
        if (newStatus === "completed") {
            hasData = true;
            if (mergeFn) {
                data = mergeFn(value, data);
            }
            else {
                data = value;
            }
            onSuccess === null || onSuccess === void 0 ? void 0 : onSuccess(data);
            cleanup();
        }
        else if (newStatus === "faulted") {
            error = value;
            onError === null || onError === void 0 ? void 0 : onError(error);
            parent === null || parent === void 0 ? void 0 : parent.onChildError(error);
            cleanup();
        }
        onUpdate === null || onUpdate === void 0 ? void 0 : onUpdate(data, error);
        emitter.emit("update", flow);
        notifyUpdate();
    }
    function iteratorDone(iterator, done, value) {
        if (!isRunning())
            return;
        if (done) {
            if (iteratorStack.length) {
                return iteratorNext(iteratorStack.pop(), value);
            }
            return statusChanged("completed", value, false);
        }
        return iteratorYield(iterator, value);
    }
    function iteratorThrow(iterator, error) {
        var _a;
        if (!isRunning())
            return;
        try {
            var _b = ((_a = iterator.throw) === null || _a === void 0 ? void 0 : _a.call(iterator, error)) || {}, done = _b.done, value = _b.value;
            // error handled
            return iteratorDone(iterator, done, value);
        }
        catch (e) {
            if (iteratorStack.length) {
                return iteratorThrow(iteratorStack.pop(), e);
            }
            // error is not handled
            return statusChanged("faulted", e, false);
        }
    }
    function interatorCall(iterator, next) {
        iteratorStack.push(iterator);
        return iteratorNext(next, undefined);
    }
    function iteratorYield(iterator, value) {
        if (!isRunning())
            return;
        cancelFn = undefined;
        if (value) {
            // yield iterator
            if (typeof value.next === "function") {
                return interatorCall(iterator, value);
            }
            // yield promise
            if (value && typeof value.then === "function") {
                cancelFn = value.cancel;
                return value.then(function (resolved) {
                    iteratorNext(iterator, resolved);
                }, function (rejected) {
                    iteratorThrow(iterator, rejected);
                });
            }
            if (typeof value === "function") {
                try {
                    var c_1 = called;
                    var ec = {
                        flow: flow,
                        controller: controller,
                        context: controller.context,
                        next: function (value) {
                            if (c_1 !== called) {
                                return;
                            }
                            iteratorNext(iterator, value);
                        },
                        fail: function (error) {
                            if (c_1 !== called) {
                                return;
                            }
                            iteratorThrow(iterator, error);
                        },
                        call: function (next) {
                            var args = [];
                            for (var _i = 1; _i < arguments.length; _i++) {
                                args[_i - 1] = arguments[_i];
                            }
                            if (c_1 !== called) {
                                return;
                            }
                            if (typeof next === "function") {
                                try {
                                    return iteratorYield(iterator, next.apply(void 0, args));
                                }
                                catch (e) {
                                    return iteratorThrow(iterator, e);
                                }
                            }
                            return interatorCall(iterator, next);
                        },
                        end: function (result) {
                            if (c_1 !== called) {
                                return;
                            }
                            iteratorStack.length = 0;
                            iteratorDone(iterator, true, result);
                        },
                    };
                    return value(ec);
                }
                catch (e) {
                    return iteratorThrow(iterator, e);
                }
            }
        }
        throw new Error("Unsupported yield " + typeof value + ": " + value);
    }
    function iteratorNext(iterator, payload) {
        if (!isRunning())
            return;
        if (childError) {
            return iteratorThrow(iterator, childError);
        }
        try {
            var _a = iterator.next(payload), done = _a.done, value = _a.value;
            return iteratorDone(iterator, done, value);
        }
        catch (e) {
            iteratorThrow(iterator, e);
        }
    }
    function isStale(triggerUpdated) {
        var now = Date.now();
        if (stale !== 0 && stale !== 1 && stale <= now) {
            // mark as stale and do not trigger updated
            stale = 1;
            if (triggerUpdated) {
                notifyUpdate();
            }
            cleanup();
        }
        return stale !== 0 && stale <= now;
    }
    var flow = {
        key: key,
        fn: fn,
        controller: controller,
        statusChanged: statusChanged,
        get expiry() {
            return expiry;
        },
        get extra() {
            return extra;
        },
        get called() {
            return called;
        },
        get completed() {
            return status === "completed";
        },
        get faulted() {
            return status === "faulted";
        },
        get running() {
            return status === "running";
        },
        get idle() {
            return status === "idle";
        },
        get hasData() {
            return hasData || false;
        },
        get current() {
            return controller.flow(key) || flow;
        },
        get parent() {
            return parent;
        },
        get previous() {
            return previous;
        },
        get data() {
            return data;
        },
        set data(value) {
            flow.update(value);
        },
        get status() {
            return status;
        },
        get error() {
            return error;
        },
        get stale() {
            return isStale(false);
        },
        setExpiry: function (value) {
            expiry = value;
        },
        setMerge: function (value) {
            mergeFn = value;
        },
        setBlock: function (value) {
            if (typeof value === "boolean") {
                blocked = value ? Date.now() + Number.MAX_VALUE : 0;
            }
            else if (value instanceof Date) {
                blocked = value.getTime();
            }
            else if (typeof value === "number") {
                blocked = value;
            }
            else {
                throw new Error("Invalid block value");
            }
        },
        setStale: function (value) {
            // cannot turn stale to unstale
            if (stale !== 0 && stale < value)
                return;
            stale = value;
            isStale(true);
        },
        get cancelled() {
            return cancelled || (parent === null || parent === void 0 ? void 0 : parent.cancelled) || false;
        },
        on: emitter.on,
        cancel: function () {
            if (cancelled)
                return flow;
            cancelled = true;
            blocked = 0;
            cancelFn === null || cancelFn === void 0 ? void 0 : cancelFn();
            emitter.emit("cancel", flow);
            cleanup();
            return flow;
        },
        partial: function (partialData, wait) {
            if (!isRunning())
                return flow;
            data = partialData;
            hasData = true;
            if (wait) {
                status = "completed";
            }
            notifyUpdate();
            return flow;
        },
        start: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            // running or finished
            if (status !== "idle" && !isStale(false) && !isExpired()) {
                return flow;
            }
            cancelFn = undefined;
            cancelled = false;
            mergeFn = undefined;
            stale = 0;
            expiry = 0;
            blocked = 0;
            called++;
            if (flow.previous) {
                controller.replaceFlow(flow.previous, flow);
            }
            emitter.emit("start", flow);
            try {
                status = "running";
                var result = fn.apply(void 0, args);
                // promise
                if (result && typeof result.then === "function") {
                    cancelFn = result.cancel;
                    statusChanged("running", undefined, false);
                    result.then(function (resolved) {
                        statusChanged("completed", resolved, false);
                    }, function (rejected) {
                        statusChanged("faulted", rejected, false);
                    });
                }
                // iterator
                else if (result && typeof result.next === "function") {
                    iteratorNext(result, undefined);
                    if (isRunning()) {
                        notifyUpdate();
                    }
                }
                else {
                    statusChanged("completed", result, false);
                }
            }
            catch (e) {
                statusChanged("faulted", e, false);
            }
            return flow;
        },
        restart: function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            if (blocked && blocked >= Date.now()) {
                return flow;
            }
            if (status === "idle" || stale) {
                return flow.start.apply(flow, args);
            }
            return (_a = createFlow({
                controller: controller,
                parent: parent,
                previous: flow,
                key: key,
                fn: fn,
                onSuccess: onSuccess,
                onError: onError,
                initData: data,
                hasData: hasData,
                extra: extra,
            })).start.apply(_a, args);
        },
        update: function (value) {
            if (status !== "running") {
                if (typeof value === "function") {
                    value = value(data);
                }
                mergeFn = undefined;
                statusChanged("completed", value, true);
            }
            return flow;
        },
        dispose: function () {
            if (disposed)
                return flow;
            disposed = true;
            cancelled = true;
            emitter.dispose();
            return flow;
        },
        onChildError: function (error) {
            childError = error;
        },
        setNext: function (value) {
            currentNext = value;
        },
        next: function (payload) {
            if (!cancelled && !isStale(true) && currentNext) {
                var next = currentNext;
                currentNext = undefined;
                status = "running";
                next === null || next === void 0 ? void 0 : next.call(null, payload);
                if (status === "running") {
                    notifyUpdate();
                }
            }
            return flow;
        },
    };
    return flow;
}
exports.createFlow = createFlow;
function makeKey(args) {
    return args.join(":");
}
exports.makeKey = makeKey;
function getKey(value) {
    if (typeof value === "function") {
        return value.flowKey || value;
    }
    if (Array.isArray(value)) {
        return makeKey(value);
    }
    return String(value);
}
exports.getKey = getKey;
function configure(fn, configs) {
    var key = ((typeof configs === "string" ? { key: configs } : configs) || {}).key;
    if (typeof key !== "undefined") {
        fn.flowKey = key;
    }
    return fn;
}
exports.configure = configure;
function createEffect(fn) {
    return fn;
}
exports.createEffect = createEffect;
function createEmitter(_a) {
    var _b = _a === void 0 ? {} : _a, wildcard = _b.wildcard, privateEventPrefix = _b.privateEventPrefix;
    var events = new Map();
    function getHandlers(event) {
        var handlers = events.get(event);
        if (!handlers) {
            handlers = [];
            events.set(event, handlers);
        }
        return handlers;
    }
    function on() {
        var args = [];
        for (var _i = 0; _i < arguments.length; _i++) {
            args[_i] = arguments[_i];
        }
        if (args.length > 1) {
            var _a = args, event_1 = _a[0], listener_1 = _a[1];
            var removes_1 = (Array.isArray(event_1) ? event_1 : [event_1]).map(function (event) {
                var handlers = getHandlers(event);
                handlers.push(listener_1);
                var active = true;
                return function () {
                    if (!active)
                        return;
                    active = false;
                    var index = handlers.indexOf(listener_1);
                    if (index !== -1)
                        handlers.splice(index, 1);
                    return listener_1;
                };
            });
            if (Array.isArray(event_1)) {
                return function () { return removes_1.forEach(function (x) { return x; }); };
            }
            return removes_1[0];
        }
        var cancel = function () { };
        return Object.assign(new Promise(function (resolve) {
            cancel = on(args[0], function (payload) {
                cancel();
                resolve(payload);
            });
        }), { cancel: cancel });
    }
    function emit(event, payload) {
        try {
            var handlers = getHandlers(event);
            handlers.slice(0).forEach(function (handler) { return handler(payload); });
        }
        finally {
            if (wildcard &&
                event !== wildcard &&
                (!privateEventPrefix ||
                    typeof event !== "string" ||
                    privateEventPrefix !== event[0])) {
                emit(wildcard, { type: event, payload: payload });
            }
        }
    }
    function dispose() {
        events.forEach(function (listeners) {
            listeners.length = 0;
        });
        events.clear();
    }
    return { emit: emit, on: on, dispose: dispose };
}
exports.createEmitter = createEmitter;
var delay = function (ms) {
    if (ms === void 0) { ms = 0; }
    var timeout;
    return Object.assign(new Promise(function (resolve) { return (timeout = setTimeout(resolve, ms)); }), {
        cancel: function () {
            clearTimeout(timeout);
        },
    });
};
exports.delay = delay;
function listenerChain(listenerFactory, register) {
    var removeListeners = [];
    var cleanup = function () {
        removeListeners.forEach(function (x) { return x(); });
    };
    register(listenerFactory(cleanup), function (x) {
        removeListeners.push(x);
        return x;
    }, cleanup);
    return cleanup;
}
exports.listenerChain = listenerChain;
function isPlainObject(value) {
    return typeof value == "object" && value.constructor == Object;
}
