import * as React from "react";
import { AnyFunc, FlowController, FlowDataInfer, FlowStatus, Flow, FlowExecutor } from "../lib";
export interface FlowProviderProps {
    controller?: FlowController;
    defaultFlow?: AnyFunc | Record<string, AnyFunc>;
    suspense?: boolean;
    errorBoundary?: boolean;
    fallback?: React.ReactNode;
}
export interface FlowHookWithoutArgs<T extends AnyFunc> {
    readonly current: Flow<T> | undefined;
    readonly data: FlowDataInfer<T> | undefined;
    readonly status: FlowStatus | "unknown";
    readonly idle: boolean;
    readonly running: boolean;
    readonly completed: boolean;
    readonly faulted: boolean;
    readonly error: Error | undefined;
    readonly cancelled: boolean;
    readonly stale: boolean;
    readonly hasData: boolean;
    readonly disabled: boolean;
    update(data: FlowDataInfer<T>): this;
    update(reducer: (prev: FlowDataInfer<T>) => FlowDataInfer<T>): this;
    next(payload?: any): this;
    start(): this;
    restart(): this;
    cancel(): this;
}
export interface FlowHook<T extends AnyFunc> extends FlowHookWithoutArgs<T> {
    start(...args: Parameters<T>): this;
    restart(...args: Parameters<T>): this;
    get(...args: Parameters<T>): FlowDataInfer<T> | undefined;
    tryGet(defaultValue: FlowDataInfer<T>, ...args: Parameters<T>): FlowDataInfer<T>;
}
export interface UseFlowOptions<T extends AnyFunc = AnyFunc> extends UseFlowOptionsWithArgs<T> {
    flow: T;
    key?: any;
}
export interface UseFlowOptionsWithArgs<T extends AnyFunc> extends UseFlowOptionsWithoutArgs<T> {
    args?: Parameters<T>;
}
export interface UseFlowOptionsWithoutArgs<T extends AnyFunc = AnyFunc> {
    defaultData?: FlowDataInfer<T>;
    suspense?: boolean;
    errorBoundary?: boolean;
    disabled?: boolean;
    onUpdated?: (flow: Flow<T>) => void;
    onCompleted?: (data: FlowDataInfer<T>) => void;
    onFaulted?: (error: Error) => void;
}
export interface PrefetchMapFn extends Function {
    <T extends AnyFunc, TResult>(key: any, flow: T, mapper: (flow: Flow) => TResult, defaultValue?: TResult): TResult | undefined;
    <TResult>(key: any, mapper: (flow: Flow) => TResult, defaultValue?: TResult): TResult | undefined;
}
export declare function useController(): FlowController;
export declare function usePrefetcher(): [FlowExecutor, PrefetchMapFn];
export interface UseFlow extends Function {
    <T extends AnyFunc>(options: UseFlowOptions<T>): FlowHook<T>;
    <T extends AnyFunc>(options: Exclude<UseFlowOptions<T>, "args">, deps: Parameters<T>): FlowHook<T>;
    <T extends AnyFunc>(key: [string, ...Parameters<T>], flow: T, options?: UseFlowOptionsWithoutArgs<T>): FlowHookWithoutArgs<T>;
    <T extends AnyFunc>(key: [string, ...Parameters<T>], flow: T, options?: UseFlowOptionsWithoutArgs<T>): FlowHookWithoutArgs<T>;
    <T extends AnyFunc>(key: string, flow: T, options?: UseFlowOptionsWithArgs<T>): FlowHook<T>;
    <T extends AnyFunc>(key: string, flow: T, deps: Parameters<T>, options?: UseFlowOptionsWithoutArgs<T>): FlowHook<T>;
    <T extends AnyFunc>(flow: T, options?: UseFlowOptionsWithArgs<T>): FlowHook<T>;
    <T extends AnyFunc>(flow: T, deps: Parameters<T>, options?: UseFlowOptionsWithoutArgs<T>): FlowHook<T>;
    (key: string | [string, ...any[]], options?: UseFlowOptionsWithoutArgs): FlowHook<AnyFunc>;
    (key: string, deps: any[], options?: UseFlowOptionsWithoutArgs): FlowHook<AnyFunc>;
}
export declare const useFlow: UseFlow;
export interface FlowComponentOptions<TProps> {
    loading?: (props: TProps) => any;
    error?: (props: TProps, error: any) => any;
}
export declare const FlowProvider: React.FC<FlowProviderProps>;
