"use strict";
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __spreadArray = (this && this.__spreadArray) || function (to, from, pack) {
    if (pack || arguments.length === 2) for (var i = 0, l = from.length, ar; i < l; i++) {
        if (ar || !(i in from)) {
            if (!ar) ar = Array.prototype.slice.call(from, 0, i);
            ar[i] = from[i];
        }
    }
    return to.concat(ar || Array.prototype.slice.call(from));
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.FlowProvider = exports.useFlow = exports.usePrefetcher = exports.useController = void 0;
var React = require("react");
var lib_1 = require("../lib");
var flowContext = React.createContext(null);
function useController() {
    return React.useContext(flowContext).controller;
}
exports.useController = useController;
function usePrefetcher() {
    var controller = useController();
    var keys = React.useRef(new Set()).current;
    var rerender = React.useState()[1];
    React.useEffect(function () {
        return controller.on(lib_1.FLOW_UPDATE_EVENT, function (flow) {
            if (!keys.has(flow.key))
                return;
            rerender({});
        });
    }, [keys, rerender, controller]);
    return React.useMemo(function () { return [
        function () {
            var _a;
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            var key = Array.isArray(args[0]) ? (0, lib_1.makeKey)(args[0]) : args[0];
            keys.add(key);
            var flow = (_a = controller).start.apply(_a, args);
            return flow;
        },
        function (key) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            var _a = typeof args[0] === "function" && typeof args[1] === "function"
                ? args
                : __spreadArray([undefined], args, true), fn = _a[0], mapper = _a[1], defaultValue = _a[2];
            var k = Array.isArray(key) ? (0, lib_1.makeKey)(key) : key;
            var flow = fn ? controller.flow(k, fn) : controller.flow(k);
            return flow ? mapper(flow) : defaultValue;
        },
    ]; }, [controller, keys]);
}
exports.usePrefetcher = usePrefetcher;
exports.useFlow = useFlowFn;
function useFlowBase(key, fn, deps, originalOptions) {
    var provider = React.useContext(flowContext);
    if (!provider) {
        throw new Error("No flow provider found");
    }
    var prependArgs = [];
    var args = originalOptions === null || originalOptions === void 0 ? void 0 : originalOptions.args;
    // no fn presents
    if (!fn) {
        if (!provider.defaultFlow) {
            throw new Error("No provider.defaultFlow presents");
        }
        // try to find fn from defaultFlow
        if (typeof provider.defaultFlow === "function") {
            fn = provider.defaultFlow;
            if (Array.isArray(key)) {
                prependArgs.push.apply(prependArgs, key);
            }
            else {
                prependArgs.push(key);
            }
        }
        else {
            if (typeof key === "undefined") {
                throw new Error("No key presents. Cannot find flow by key");
            }
            if (Array.isArray(key)) {
                fn = provider.defaultFlow[key[0]];
            }
            else {
                fn = provider.defaultFlow[key];
            }
        }
    }
    if (typeof key === "undefined") {
        key = fn;
    }
    if (Array.isArray(key)) {
        args = key.slice(1);
    }
    var _a = originalOptions || {}, _b = _a.suspense, suspense = _b === void 0 ? provider.suspense : _b, _c = _a.errorBoundary, errorBoundary = _c === void 0 ? provider.errorBoundary : _c, options = __rest(_a, ["suspense", "errorBoundary"]);
    var fixedArgs = !!args;
    var rerender = React.useState()[1];
    var optionsRef = React.useRef({ prependArgs: prependArgs });
    var renderingRef = React.useRef(false);
    var flowRef = React.useRef();
    var firstRunRef = React.useRef(true);
    var _d = React.useMemo(function () {
        return createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs);
    }, [suspense, errorBoundary, fixedArgs]), flowHook = _d.flowHook, handleSuspeseAndErrorBoundary = _d.handleSuspeseAndErrorBoundary;
    // update refs
    renderingRef.current = true;
    optionsRef.current = __assign(__assign({}, options), { args: args, suspense: suspense, errorBoundary: errorBoundary, prependArgs: prependArgs });
    // make sure removed flow does not cause an error
    flowRef.current =
        provider.controller.flow(key, fn) || flowRef.current;
    React.useLayoutEffect(function () {
        var _a;
        renderingRef.current = false;
        var status = (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status;
        return provider.controller.on(lib_1.FLOW_UPDATE_EVENT, function (flow) {
            var _a, _b, _c, _d, _e, _f;
            if (optionsRef.current.disabled)
                return;
            if (flow.key === key) {
                rerender({});
                if (flow.status !== status) {
                    (_b = (_a = optionsRef.current).onUpdated) === null || _b === void 0 ? void 0 : _b.call(_a, flow);
                    if (flow.completed) {
                        (_d = (_c = optionsRef.current).onCompleted) === null || _d === void 0 ? void 0 : _d.call(_c, flow.data);
                    }
                    else if (flow.faulted) {
                        (_f = (_e = optionsRef.current).onFaulted) === null || _f === void 0 ? void 0 : _f.call(_e, flow.error);
                    }
                }
            }
        });
    });
    handleSuspeseAndErrorBoundary();
    if (deps) {
        var firstRun_1 = firstRunRef.current;
        firstRunRef.current = false;
        React.useMemo.call(null, function () { return (firstRun_1 ? flowHook.start.apply(flowHook, deps) : flowHook.restart.apply(flowHook, deps)); }, deps);
    }
    return flowHook;
}
function useFlowFn() {
    var args = [];
    for (var _i = 0; _i < arguments.length; _i++) {
        args[_i] = arguments[_i];
    }
    // overload #1, #2
    if (typeof args[0] === "object" && !Array.isArray(args[0])) {
        var options_1 = args[0];
        return useFlowBase(options_1.key, options_1.flow, args[1], options_1);
    }
    // overload #7, #8
    if (typeof args[0] === "function") {
        // overload #8 (flow, deps, options)
        if (Array.isArray(args[1])) {
            var flow_1 = args[0], deps = args[1], options_2 = args[2];
            return useFlowBase(undefined, flow_1, deps, options_2);
        }
        var flow_2 = args[0], options_3 = args[1];
        return useFlowBase(undefined, flow_2, undefined, options_3);
    }
    if (Array.isArray(args[2])) {
        var key_1 = args[0], flow_3 = args[1], deps = args[2], options_4 = args[3];
        return useFlowBase(key_1, flow_3, deps, options_4);
    }
    var key = args[0], flow = args[1], options = args[2];
    return useFlowBase(key, flow, undefined, options);
}
function createFlowHook(flowRef, renderingRef, optionsRef, fixedArgs) {
    function handleSuspeseAndErrorBoundary() {
        if (!flowRef.current || !renderingRef.current)
            return;
        if (optionsRef.current.errorBoundary &&
            flowRef.current.status === "faulted") {
            throw flowRef.current.error;
        }
        if (optionsRef.current.suspense && flowRef.current.status === "running") {
            throw new Promise(function (resolve) {
                var _a;
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.on("update", resolve);
            });
        }
    }
    function getArgs(inputArgs) {
        var prependArgs = optionsRef.current.prependArgs || [];
        var defaultArgs = optionsRef.current.args;
        if (!defaultArgs) {
            return __spreadArray(__spreadArray([], prependArgs, true), inputArgs, true);
        }
        return __spreadArray(__spreadArray(__spreadArray([], prependArgs, true), inputArgs, true), defaultArgs.slice(inputArgs.length), true);
    }
    function run(type, args) {
        var _a;
        if (optionsRef.current.disabled) {
            return flowHook;
        }
        if (args.length && fixedArgs) {
            throw new Error("Passing arguments to fixed args flow is not allowed. Use the overload useFlow(key, flow, { args: [] }) instead");
        }
        (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a[type].apply(_a, getArgs(args));
        handleSuspeseAndErrorBoundary();
        return flowHook;
    }
    var flowHook = {
        get hasData() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData) || false;
        },
        get status() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.status) || "unknown";
        },
        get error() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.error;
        },
        get running() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.running) || false;
        },
        get completed() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.completed) || false;
        },
        get faulted() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.faulted) || false;
        },
        get idle() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.idle) || false;
        },
        get data() {
            var _a, _b;
            if (!((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.hasData)) {
                return optionsRef.current.defaultData;
            }
            return (_b = flowRef.current) === null || _b === void 0 ? void 0 : _b.data;
        },
        get disabled() {
            return optionsRef.current.disabled || false;
        },
        get current() {
            var _a;
            return (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.current;
        },
        get cancelled() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancelled) || false;
        },
        get stale() {
            var _a;
            return ((_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.stale) || false;
        },
        update: function (data) {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.update(data);
            }
            return flowHook;
        },
        get: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return flowHook.start.apply(flowHook, args).data;
        },
        tryGet: function (defaultValue) {
            var args = [];
            for (var _i = 1; _i < arguments.length; _i++) {
                args[_i - 1] = arguments[_i];
            }
            var _a = {
                result: flowHook.start.apply(flowHook, args).data,
            }.result, result = _a === void 0 ? defaultValue : _a;
            return result;
        },
        start: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("start", args);
        },
        restart: function () {
            var args = [];
            for (var _i = 0; _i < arguments.length; _i++) {
                args[_i] = arguments[_i];
            }
            return run("restart", args);
        },
        next: function (payload) {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.next(payload);
            }
            return flowHook;
        },
        cancel: function () {
            var _a;
            if (!optionsRef.current.disabled) {
                (_a = flowRef.current) === null || _a === void 0 ? void 0 : _a.cancel();
            }
            return flowHook;
        },
    };
    return { flowHook: flowHook, handleSuspeseAndErrorBoundary: handleSuspeseAndErrorBoundary };
}
var FlowProvider = function (props) {
    var parentProvider = React.useContext(flowContext);
    var rerender = React.useState()[1];
    var _a = props.controller, controller = _a === void 0 ? parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.controller : _a, _b = props.suspense, suspense = _b === void 0 ? (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.suspense) || false : _b, _c = props.errorBoundary, errorBoundary = _c === void 0 ? (parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.errorBoundary) || false : _c, _d = props.defaultFlow, defaultFlow = _d === void 0 ? parentProvider === null || parentProvider === void 0 ? void 0 : parentProvider.defaultFlow : _d, fallback = props.fallback, children = props.children;
    var value = React.useMemo(function () { return ({ controller: controller, suspense: suspense, errorBoundary: errorBoundary, defaultFlow: defaultFlow }); }, [controller, suspense, errorBoundary, defaultFlow]);
    if (!controller) {
        throw new Error("No controller provided");
    }
    if (controller.error) {
        throw controller.error;
    }
    if (!controller.ready) {
        if (suspense)
            throw controller.promise;
        controller.promise.finally(function () { return rerender({}); });
        return React.createElement(React.Fragment, {}, fallback);
    }
    return React.createElement(flowContext.Provider, {
        value: value,
        children: children,
    });
};
exports.FlowProvider = FlowProvider;
